(* ::Package:: *)

(* ::Text:: *)
(*(* riemannv2.m *)*)
(**)
(*(*Mathematica Version 11.1.1.0. *)*)
(*(* Package Version 2. *)*)
(**)
(*(* Author: Rolf Sulanke *)*)
(**)
(*(* First version:Started February 29, 2016. Finished  May 30, 2016;*)
(*v2: Started April 18, 2017 .Finished June 10, 2017.*)*)
(**)
(*(* This package contains tools applied in notebooks about *)
(*pseudo-Riemannian geometry or linear connections in n dimensions. *)*)
(**)
(*(* Hint:s. *)
(**)
(*1. Be sure, that besides of this package also the packages  euvecv2.m,*)
(*tensalgv3.m,  and  neuvecv2.m are contained in the working directory.*)
(**)
(*2. After importing this package define the number*)
(*dim (= dimension: number of independent variables)) and ind (the index *)
(*of the pseudo-Riemannian metric) as constants *)
(*within the Global context of the notebook! *)
(*These numbers may be changed later in the notebook, if wanted. If they are defined before *)
(*importing the notebook they will be fixed for all the functions introduced in the packages! *)*)
(**)
(*(*Copyright:  Free software:This and the accompanying packages and notebookss are  public. Authors who intend to publish a changed or completed version of them should do it under their own names with the condition that they cite the original notebook with the Internet address or other source where they got it. I am not responsible for errors or damages originated by the use of the procedures contained in my notebooks or packages; everybody who applies them should test carefully whether they are appropriate for his purposes.*)*)
(**)
(*(* Changes.*)
(*17.05.02. smf replaced by imm.*)
(*17.05.02. RGposk,rgposk entered.*)
(*17.05.11 tensorfield added.*)
(*17.05.13  vv corrected.*)
(*17.05.18 cnf corrected.*)
(**)*)


(* ::Section:: *)
(*Context "riemann`"*)


BeginPackage["riemannv2`",{"tensalgv3`", "euvecv2`","neuvecv2`"}]


(* ::Section:: *)
(* The usages *)


imm::usage = "smf[x,...] =imm[f][x,...] is the local coordinate representation of an unspecified smooth map 
 of an m-dimensional parameter manifold into an n-dimensional smooth manifold, n = dim must 
be  defined in the global context. It is supposed that the rank rk of the Jacobian matrix 
of smf is maximal, i.e. rk[d[smf][x]]= Min[m,n], for almost all points x={xi}.
Points x for which the rank is not maximal are called singular, the others 
are regular. Such a map is said to be an immersion, if m\[LessEqual] n, and a submersion, 
if m\[GreaterEqual]n. In the case m = n the map smf is supposed to be a local diffeomorphism.";

vv::usage = "vv[a][x1,...,xm] is a vector field on the m-manifold with 
coordinates  {x1,...,xm} with components a[x1,...,xm][i], i = 1,...,m.";

d::usage = "d[f][u,...] or d[f][{u,...}]is the differential of the function f[u,...] of several variables 
with values in an n-dimensional manifold or space. For the variables u, ... the differentials 
d[u][u,v,...], ... yield the standard cobasis in the cotangent space of the parameter 
manifold at the point {u, ...}. d[f][u,..].vl is the value of the differential d[f][u,...] 
for the tangent vector vl.";

jac::usage = "jac[f][x1,...] or jac[f][{x1,...}] is the transposed Jacobian matrix of a map f of the m-dimensional 
parameter manifold with coordinates {x1,...,xm) into the n-dimensional space if n>1, and d[f][x] if n==1.";

tv::usage = "tv[f][i][u,v,...] is the tangent vector of the i-th  parameter line of 
the immersion f[u,v,...] at the point {u, v,...}. If this point is regular these 
vectors for i = 1,...,Length[{u,v,...}] form a basis of the tangent space 
image[d[f][u,v,...]].";

vv::usage = "vv[a][x1,...,xm] is a vector field on the m-manifold with 
coordinates  {x1,...,xm} with components a[i][x1,...,xm], i = 1,...,m.";

tvf::usage = "tvf[f][a][u,v,...] is the tangent vector field on the 
m-surface f[u,v...] with coordinates a[i][u,v,...] with respect to the 
standard basis tv[f][i][u,v,...].";



(* ::Subsection:: *)
(*Manifolds*)


var::usage = "var[x] is the sequence of the coordinates x[1],...,x[n] in a chart, n = dim.";

\[Delta]::usage = "\[Delta][i,j] is the short form of KroneckerDelta[i,j].";

identity::usage = "identity[u1,...,um] ={u1,...,um} is the identity function on the 
manifold U with coordinates u1,...,um.";

tx\[Iota]::usage = "The local vector field tx\[Iota][i][x[1],...,x[n]] yields the i-th vector of the standard basis 
corresponding to a chart \[Phi]\[Iota] of the manifold \!\(\*SuperscriptBox[\(M\), \(n\)]\) at the points x 
with coordinates {x[1],...,x[n]}."; 

der::usage = "der[vlist,h][x1,...,xn] is the directional derivative of the function h[x1,...,xn] in 
direction of the tangent vector with coordinates vlist, calculated in local coordinates 
with the local representative of h.";

comm::usage ="comm[X,Y][x1,...,xn] yields the commutator of the local vector fields X[x1,...,xn],Y[x1,...,xn] 
on a differentiable manifold.";

tensorfield::usage ="tensorfield[a,p,q][x1,..,xn] yields a tensor of type (p,q)at the point with 
coordinates {x1,..,xn} whose tensor coordinates are a[ii,jj][x1,..,xn].";


(* ::Subsubsection:: *)
(*(* Exterior differential dd *)*)


dd::usage = "dd[ta][x,...] yields the exterior differential of the p-form ta[x,...];
remember:a 0-form ta[x...] must be a function with the Attribute NumericFunction; 
for antisymmetric covariant tensors ta[x,...] of degree p the operator dd returns an
antisymmetric covarient tensor of degree p+1, and for an antisymmetric Array ta[x,...] 
it returns the derived antisymmetric Array; all dimensions of the Array ta[x,...] 
must equal the number of the variables x,...";


(* ::Subsection:: *)
(*Linear connections*)


\[CapitalGamma]::usage = "\[CapitalGamma][k,i,j][x,y,...] denotes the \[CapitalGamma]-symbols of a general unspecified linear connection.";

\[CapitalGamma]g::usage = "\[CapitalGamma]g[g][k,i,j][x,y,...] are the Christoffel symbols of the second kind of the 
pseudo-Riemannian metric {g[x,y,...][i,j]} in local coordinates x,y,...";

cnf::usage = "cnf[\[CapitalGamma]][i,j][x][v]evaluates the connection form with indices (i,j) of the linear 
connection \[CapitalGamma] for tangent vectors v at the point x of the manifold. x and v must be entered by 
their coordinates.";

nabla0::usage ="nabla0[f][x1,...,xn] is the tensor field of type (1,0) corresponting to the differential of the 
function f[x1,...,xn] of the local coordinates (x1,...,xn),"; 

nabla1::usage = "nabla1[\[CapitalGamma],Y][x1,...,xn] is the  tensor field of type (1,1), named the gradient tensor, 
of the vector field Y[x1,...,xn] in local coordinates (x1,...,xn)."; 

nab1::usage = "nab1[\[CapitalGamma],Y][x1,...,xn][h,i] is the (h,i)-coordinate of the gradient tensor field 
nabla1[\[CapitalGamma],Y][x1,...,xn] of the vector field Y[x1,...,xn] in local coordinates (x1,...,xn).";

nab::usage = "nab[\[CapitalGamma], th][x][ii,jj] is the coordinate of the gradient tensor of the tensor th[x] 
with respect to the linear connection \[CapitalGamma] corresponding to the index tuples ii, jj.";

nabla::usage ="nabla[\[CapitalGamma], th][x] yields the gradient tensor of the tensor field th[x] with respect 
to the linear connection \[CapitalGamma].";

absD::usage = "absD[\[CapitalGamma], ta][x][X][ii,jj] yields the coordinate of the absolute differential of the 
tensor field ta[x] at the tangent vector X or vector field X[x] with respect to the linear 
connection  \[CapitalGamma] corresponding to the index tuples ii,jj.";

tensorD::usage = "tensorD[\[CapitalGamma], ta][x] is the absolute differential of the tensor field ta[x] 
with respect to the linear connection \[CapitalGamma]; tensorD[\[CapitalGamma],ta][x][X] evaluates the absolute differential 
at the tangent vector X or the vector field X[x].";

absD1::usage = "absD1[\[CapitalGamma], ta][cu,t][ii,jj] yields the coordinate of the absolute differential of the 
tensor field ta[t] for the tangent vector cu'[t] along the curve cu[t] with respect to 
the linear connection  \[CapitalGamma] corresponding to the index tuplesii,jj.";

tensorD1::usage = "tensorD1[\[CapitalGamma], ta][cu,t] is the absolute differential of the tensor field ta[t] along 
the curve cu[t] with respect to the linear connection \[CapitalGamma] evaluated for the tangent vector cu'[t] of the curve.";

tors::usage = "tors[\[CapitalGamma]}[x1,...,xn] is the torsion tensor field of a linear connection \[CapitalGamma] on a 
n-dimensional manifold.";

RC::usage ="RC[\[CapitalGamma]][x1,...,xn][k,i,h,j]is the component of the curvature tensor of the linear 
connection \[CapitalGamma] with indices k (contravariant) and h,i,j (covariant) in a n-dimensional smooth manifold.";

RCtensor::usage = "RCtensor[\[CapitalGamma]][x1,...,xn] is the curvature tensor (of type (3, 1)) of the linear 
connection \[CapitalGamma] on a smooth n-dimensional manifold. Its coordinates are RC[\[CapitalGamma]][x1,...,xn][k,i,h,j].";

Riccitensor::usage = "Riccitensor[\[CapitalGamma]][x1,...,xn] is the Ricci curvature tensor of the linear connection \[CapitalGamma].";

rct::usage ="rct[\[CapitalGamma]][x1,...,xn][i,j] is the (i,j)-coordinate of the Ricci curvature tensor of the linear connection \[CapitalGamma].";


(* ::Subsection:: *)
(*Riemannian spaces*)


RG::usage = "RG[gg:g][x1,...] is the matrix of the elements gg[x1,...][i,j] of a symmetric matrix 
field on the manifold with coordinstes {x1,...}; the default argument gg=g gives a general, 
unspecified symmetric matrix field.";

rg::usage = "rg[gg_:g][x__][i,j] is the (i,j)-element of the symmetric matrix RG[gg][x].";

(*rg::usage = "rg[gg_:g][x__][i,j] is the (i,j)-element of the symmetric matrix RG[{x,..},gg].";
 
genRG::usage = "genRG[x1,...,xn]= RG[{x1,...}] is the matrix of the unspecified coordinates 
g[x1,...][i,j] of a symmetric matrix field on the manifold with coordinstes {x,...}.";

specRG::usage = "specRG[h][x1,...xn]= RG[{x1,...xn},h] is the symmetric matrix of the 
coordinates h[x1,...][i,j][x1,...] of a symmetric matrix field on the manifold with 
coordinstes {x1,...,xn}."; *)

RGtensor::usage = "RGtensor[gg][x,y,...] is the covariant symmetric tensor field of 
degree 2 whose coordinate Array is RG[gg][x,y,...].";

INV::usage ="If m[u,v,...] is an invertible square matrix of functionsf[u,v,...][i,j] 
then INV[m][u,v..] is the inverse matrix.";

inv::usage = "inv[m][u,v,...][i,j] is the (i,j)-element of the matrix INV[m][u,v,...].";

eucl::usage = "eucl[x1,...,xn][i,j] =\[Delta][i,j]are the coordinates of the n-dimensional 
Euclidean metric tensor in orthonormal coordinates.";

EUCL::usage =  "EUCL[x1,x2,...,xn]is the matrix of the coordinates of the 
n-diemensional Euclidean metric with respect to orthonormal Cartesian coordinates.";

lor::usage = "lor[i,j,k:ind] is the standard pseudo-orthonormal form of the metric 
of the pseudo-Euclidean vector space of index k with default k=ind;i,j = 1,...,dim.";

lorg::usage = "lorg[x1,...,xn][i,j,k:ind] is the standard pseudo-orthonormal form of the 
pseudo-Riemannian metric of the pseudo-Euclidean space of index k with default 
k=ind;i,j = 1,...,n=dim.";


 RGposk::usage = "RGposk[r][s1,s2,...,sn]  is the matrix of the coordinates of the 
n-dimensional Riemannian metric of positive constant curvature 1/r^2 with respect 
to the spherical coordinates (s1,...,sn) used in the definition of unit.";

rgposk::usage = "rgposk[r][s1,s2,...,sn][i,j] is the (i,j)-coordinate of 
RGposk[r][s1,s2,...,sn].";

RGT::usage = "RGT[f][u,v,...]is the matrix of the coordinates of the first fundamental 
form of the immersion f[u,v,...] into the n-dimensional Euclidean space: Its Riemannian inner metric.";

rgt::usage = "rgt[f][x,y,...][i,j]is the (i,j)-coordinate of the inner Riemannian 
metric of the immersion f[x,y,...] into an Euclidean space.";

dVol::usage = "dVol[\[Gamma]\[Gamma]][u,v,...] is the volume element of the pseudo-Riemannian manifold 
with scalar product \[Gamma]\[Gamma]. See also dV.";

dV::usage = "dV[f][u,v,...] is the volume element of the immersion f[u,v,...] in 
the Euclidean Space.. See also dVol.";

chr::usage = "chr[i,j,k,g:rg[]][x,y,...] are the Christoffel symbols of the 
first kind of the pseudo-Riemannian metric {g[i,j][x,y,...]} in local 
coordinates x,y,... The default g is the unspecified symmetric gr.";

\[CapitalGamma]g::usage = "\[CapitalGamma]g[g:rg[]][k,i,j][x,y,...] are the Christoffel symbols of the second kind 
of the pseudo-Riemannian metric {g[x,y,...][i,j]} in local coordinates x,y,...  
\[CapitalGamma]g[ ][k,i,j][x,y,...] yields the Christoffel symbpls of the default unspecified metric rg.
See also \[CapitalGamma].";

\[CapitalGamma]f::usage ="\[CapitalGamma]f[ff][h,i,j][u,v,...], h,i,j = 1,...,Length[{u,v,...}], are the Christoffel 
symbols of the second kind of the Riemannian inner metric  of the immersion ff[u,v,...] 
into the Euclidean n-space.";

\[CapitalGamma]0::usage = "\[CapitalGamma]0[[i,j,k]x,...] (= 0 in Cartesian coordinates) are the \[CapitalGamma]-symbols of the
 flat linear connection.";

covRCtensor::usage = "covRCtensor[gg:rg[]][x,y,...] is the covariant curvature 
tensor field of the pseudo-Riemannian space with metric tensor {gg[x,y,...][i,j]}.";

covrc::usage = "covrc[[gg:rg[]][k,h,i,j][x,y,...] are the coordinates of the covariant 
curvature tensor covRCtensor[gg:rg[]][x,y,...].";

sectionalcurvature::usage = "sectionalcurvature[gg:rg[]][x,y,...][X, Y] calculates 
the sectional curvature for the tangent plane span[X,Y] at the point {x,y,...} of 
a pseudo-Riemannian manifold with metric tensor gg.";

tensorscpr::usage = "For tensors ta,tb of the same type (p, q) over a finite dimensional 
vector space with a scalar product g with coordinates g[i,j] the function 
tensorscpr[g][ta,tb] is the value of the tensor scalar product for the tensors ta,tb.";

tensorfieldscpr::usage = "tensorfieldscpr[gg:rg[]][ta,tb][x,y,...] calculates the 
tensor scalar product of the tensor fields ta, tb of the same type
(p, q) over a pseudo-Riemannian manifold with the metric gg[x,y,..][i,j] 
and coordinates x,y,...";

spher::usage = "spher[r][s1,...,sn] is a parameter representation of the n-dimensional 
sphere of radius r with center at the origin of the (n+1)-dimensional Euclidean space.";

sph::usage  = "sph[r][x1,...,xn] is a parameter representation of the n-sphere of 
radius r with center at the origin  of the (n+1)-dimensional Euclidean space, defined 
by the inverse stereographic projection.";

RGpos::usage = "RGpos[r][s1,...,sn] is the matrix of the coordinates of the inner 
Riemannian metric (first fundamental form) of the n-dimensional sphere of 
radius r>0 in spherical cordinates s1,...,sn.";

RGcnfpos::usage ="RGcnfpos[r][x1,...,xn] is the matrix of the coordinates of 
the Riemannian metric of positive curvature 1/r^2 in the conformal model.";

rgcnfpos::usage="rgcnfpos[r][i,j][x1,...,xn] is the (i,j)-coordinate of the 
Riemannian metric tensor of constant positive curvature 1/r^2 in the 
conformal model.";

hypsp::usage = "hypsp[r][s1,...,sn] is a parameter representation of 
the pseudo-Euclidean model of the n-dimensional hyperbolic spacee 
with curvature -1/r^2.";

RGhypsp::usage = "RGhypsp[r][s1,s2,...,sn] is the metric tensor 
of the n-dimensional hyperbolic space of constant curvature 
-1/r^2 given by its components with respect to the coordinates o
f the pseudo-Euclidean model.";

rghypsp::usage = "rghypsp[r][s1,s2,...,sn][i,j] is the (i,j)-
coordinate of RGhypsp[r][s1,s2,...,sn].";

hypcnf::usage = "hypcnf[r][s1,...,sn] for s1^2+...+sn^2<1 is a 
parameter representation of the conformal model of the n-dimensional 
hyperbolic spacee with curvature -1/r^2.";

RGhypcnf::usage = "RGhypcnf[r][s1,s2,...,sn] is the metric tensor of the 
n-dimensional hyperbolic space of constant curvature -1/r^2 given by 
its components with respect to the coordinates of the spherical conformal model.";

rghypcnf::usage = "rghypcnf[r][s1,s2,...,sn][i,j] for s1^2+...+sn^2<1 is 
the (i,j)-coordinate of RGhypcnf[r][s1,s2,...,sn].";

RGneg::usage = "RGneg[r][s1,s2,...,sn] for s1^2+...+sn^2<1 is the metric 
tensor of the n-dimensional hyperbolic space of constant curvature -1/r^2 
in its n-disk conformal model.";

rgneg::usage = "rgneg[r][s1,s2,...,sn][i,j] for s1^2+...+sn^2<1 is 
the (i,j)-coordinate of RGneg[r][s1,s2,...,sn].";

X\[Iota]::usage = "X\[Iota][v][x[1],...,x[n]] is a local vector field defined 
on a chart domain  \!\(\*SubscriptBox[\(U\), \(\[Iota]\)]\) with vector coordinates given by the n-vector v.";


(* ::Section:: *)
(*Definitions*)


Begin["`Private`"] 

dimm := Global`dim; 

indd := Global`ind;

g:= Global`g;

(* psp := neuvecv2`pssp 
*)


(* ::Subsubsection:: *)
(*Manifolds*)


\[Delta][i_,j_] := KroneckerDelta[i,j];

var[x_,dd_:dimm] :=vec[x,dd]/.List-> Sequence;

imm[f_][x__] :=Table[f[i][x],{i,dimm}];

tx\[Iota][i_][x__] :=If[(i>=1)&&(i<=Length[{x}])&&IntegerQ[i], stb[i,Length[{x}]],
Print["Index ",i," invalid!"];Return[]];

X\[Iota][v_][x__]:= If[Length[v] == Length[{x}],Sum[v[[i]]*tx\[Iota][i][x],{i,Length[{x}]}],
Print["Arguments invalid!"];Return[]];

identity:={##}&;

vv[a_]:=Table[a[i][##],{i,Length[{##}]}]&;

tvf[ff_][a_]:= Sum[tv[i][ff][##]*a[i][##],{i,Length[{##}]}]&;

der[v_,ff_][x__]:=
If[(Head[v]===List)&&Length[v]==Length[{x}],Return[(d[ff][x]).v],
Print["Invalid arguments?"]; Return[]];

comm[v_,w_][x__] := If[(Length[v[x]]== Length[{x}])&&(Length[w[x]]== Length[{x}]),
Grad[w[x],{x}].v[x] -Grad[v[x],{x}].w[x], Print["Invalid arguments!"];Return[]];

Attributes[d]={Listable};
d[f_][x___]:= Module[{m=Length[{x}], u,v,i},
If[m==0,Print["Function ",f," needs arguments!"]; Return[]];
If[MemberQ[{x},f], Return[D[{x},f]]];
If[(m==1)&&Not[Head[x]===List],Return[D[f[v],v]/. v->x]];
If[m>=2,For[i=1,i<=m,i++,If[Head[{x}[[i]]]==List,
Print["Argument ",{x}[[i]]," is not a variable!"]; Return[]]]];
(*u=Table[v[i],{i,m}];
D[f[u/.List->Sequence],{u}]/.Table[v[i]->{x}[[i]],{i,m} ]];*)
D[f[x],{{x}}]]

d[f_][{x__}]:= Module[{ff,n = Length[{x}],y},
If[MemberQ[{x},f], Return[D[{x},f]]];
ff[y__]:=f[{y}];
d[ff][x]];

jac[f_][x__] := If[(Length[{f[x]}]==1)&&Not[Head[f[x]]=== List],d[f][x],Transpose[d[f][x]]];

(* Standard tangent basis *)

tv[f_][i_][u__]:= If[(Head[f[u]]== List)&& (Length[f[u]]>= Length[Flatten[{u}]])&&
 MemberQ[Range[ Length[Flatten[{u}]]],i],If[Length[{u}]==1,Return[d[f][u]],Return[Transpose[d[f][u]][[i]]]],
Print["Invalid arguments!"],Print["Invalid arguments!"]];

(* Tensor fields *)

tensorfield[a_,p_,q_][x__]:= Module[{bb,b,ii,jj,n=Length[{x}]},
If[(p==0)&&(q==0),b[{},{}]= a[x],b[ii_,jj_]:= a[ii,jj][x]];
bb=tensor[b,p,q,n];
bb];



(* ::Subsection:: *)
(*(* Exterior differential dd *)*)


dd[aa_][x__]:= Module[{ts=aa[x],ttp,he=Head[aa[x]], pp,di= Length[{x}],Arr,diff,dts},
If[Not[(he=== tensor)||(he=== List)]&&MemberQ[Attributes[aa],NumericFunction],
Return[d[aa][x]]];
If[he=== tensor,
ttp = type[ts];pp = ttp[[2,1]];
If[Not[ttp[[2,2]]== 0],Print["Tensor isn't covariant!"];Return[]];
If[Not[ttp[[3]]==di],Print["Dimensions differ!"];Return[]];
If[pp== 0,ts =aa[x][{},{}];Goto[diff]]];
If[he=== tensor,ts=tensorToArray[ts]; Goto[diff]];
Label[Arr];
pp=TensorRank[ts];
If[(pp==1)&&(Length[ Flatten[ts]]==1),pp=0;ts=Flatten[ts][[1]];Goto[diff]];
If[Not[TensorDimensions[ts][[1]]==di],Print["Dimensions differ!"];Return[]];
If[Not[TensorDimensions[ts]== Table[di,{pp}]],Print["Invalid Array!"];Return[]];
Label[diff];
If[(pp>1)&&(Not[(TensorSymmetry[ts] === Antisymmetric[Range[pp]])
||(TensorSymmetry[ts] === ZeroSymmetric[{}])]),
Print["Tensor ",aa[x]," is not antisymmetric!"];Return[]];
dts=((-1)^pp)*Simplify[(pp+1)*Normal[Symmetrize[D[ts,{{x}}],Antisymmetric[Range[pp+1]]]]];
If[(he===tensor)&&(pp>=di), Return[nulltensor[pp+1,0,di]]];
If[he===tensor, Return[ArrayTotensor[dts,pp+1]]];
dts];

dd/: dd[h1_+h2_][u__]:= dd[h1][u] +dd[h2][u];


(* ::Subsection:: *)
(*Linear connections*)


(* ::Subsubsection:: *)
(*Absolute differential*)


cnf[\[Gamma]_][i_,j_][x__][v_]:=If[Not[MemberQ[Range[Length[{x}]],i]&&MemberQ[Range[Length[{x}]],j]], 
Print["Invalid indices ", i, ", ", j, " !"];Return[],
If[(Head[v]===List)&&(Length[{x}] == Length[v]),Sum[\[Gamma][j,h,i][x]*v[[h]],{h,Length[{x}]}],
Print["Invalid argument ", v, " !"];Return[]]];

nabla0[f_][x__]:=covecTotensor[d[f][x]];

nabla1[\[CapitalGamma]_,zz_][x__] := ArrayTotensor[Table[D[zz[x][[i]],{x}[[h]]]+ Sum[zz[x][[k]]*
\[CapitalGamma][i,h,k][x],{k,1,Length[{x}]}],{h, 1,Length[{x}]},{i,1,Length[{x}]}],1];

nab1[\[CapitalGamma]_,v_][x__][h_,i_] := D[v[x][[i]],{x}[[h]]] +Sum [v[x][[l]]*\[CapitalGamma][i,h,l][x],{l,Length[{x}]}];

nab[\[CapitalGamma]_, tens_][x__][ii_,ji_]:= Module[{di=Length[{x}],pt,qt},
(* Syntax control *)
If[Head[tens[x]]===tensor, pt = type[tens[x]][[2,1]];qt= type[tens[x]][[2,2]],
Print[tens[x]," is not a tensor!"];Return[]];
If[Not[MemberQ[Tuples[Range[di],pt+1],ii]],Print["Invalid argument ",ii;, " !"];Return[]];
If[Not[MemberQ[Tuples[Range[di],qt],ji]],Print["Invalid argument ",ji, " !"];Return[]];
If[Not[di == type[tens[x]][[3]]], 
Print["Tensor dimension ",type[tens[x]][[3]], " is not the manifold dimension", di," !"]; Return[]];
absD[\[CapitalGamma], tens][x][stb[ii[[pt+1]],di]][ii[[1;;pt]],ji]];

nabla[\[CapitalGamma]_, tens_][x__]:= Module[{di=Length[{x}],pt,qt},
(* Syntax control *)
If[Head[tens[x]]===tensor, pt = type[tens[x]][[2,1]];qt= type[tens[x]][[2,2]],
Print[tens[x]," is not a tensor!"];Return[]];
If[Not[di == type[tens[x]][[3]]], 
Print["Tensor dimension ",type[tens[x]][[3]], " is not the manifold dimension", di," !"]; Return[]];
(* The definition *)
tensor[nab[\[CapitalGamma], tens][x],pt+1,qt,di]];

absD[\[CapitalGamma]_,tens_][x__][v_][ii_,ji_]:= Module[{di=Length[{x}],pt,qt,wv,dt,dti,dtj,i,j,s,l},
(* Syntax control *)
If[Head[tens[x]]===tensor, pt = type[tens[x]][[2,1]];qt= type[tens[x]][[2,2]],
Print[tens[x]," is not a tensor!"];Return[]];
If[Not[MemberQ[Tuples[Range[di],pt],ii]],Print["Invalid argument ",ii;, " !"];Return[]];
If[Not[MemberQ[Tuples[Range[di],qt],ji]],Print["Invalid argument ",ji, " !"];Return[]];
If[Not[di == type[tens[x]][[3]]], 
Print["Tensor dimension ",type[tens[x]][[3]], " is not the manifold dimension", di," !"]; Return[]];
If[(Head[v] === List)&&Length[v] == di,
wv[i_,j_]:=Sum[\[CapitalGamma][j,s,i][x]*v[[s]],{s,Length[{x}]}],Print["Invalid argument ", v, " !"];Return[]];
(* The components *)
dt = Sum[D[tens[x][ii,ji],{x}[[s]]]*v[[s]],{s,di}];
dti =Sum[Sum[tens[x][Join[ii[[1;;l-1]],{s},ii[[l+1;;]]],ji]*wv[ii[[l]],s],{s,di}],{l,pt}];
dtj = Sum[Sum[tens[x][ii,Join[ji[[1;;l-1]],{s},ji[[l+1;;]]]]*wv[s,ji[[l]]],{s,di}],{l,qt}]; 
dt-dti+dtj];

tensorD[\[CapitalGamma]_,tens_][x__][v_]:= Module[{di=Length[{x}],pt,qt},
(* Syntax control *)
If[Head[tens[x]]===tensor, pt = type[tens[x]][[2,1]];qt= type[tens[x]][[2,2]],
Print[tens[x]," is not a tensor!"];Return[]];
If[Not[(Head[v] === List)&&Length[v] == di],
Print["Invalid argument ", v, " !"];Return[]];
tensor[absD[\[CapitalGamma],tens][x][v],pt,qt,di]];


(* ::Subsubsection:: *)
(*The absolut derivation of a tensor field along a curve*)


absD1[\[CapitalGamma]_,tens_][cu_,t_][ii_,ji_]:= Module[{di,pt,qt,wv,dt,dti,dtj,i,j,s,l,v,xt},
(* Syntax control *)
If[Head[cu[t]]===List, di = Length[cu[t]],Print[cu[t]," is not a curve!"];Return[]];
If[Head[tens[t]]===tensor, pt = type[tens[t]][[2,1]];qt= type[tens[t]][[2,2]],
Print[tens[t]," is not a tensor!"];Return[]];
If[Not[di == type[tens[t]][[3]]], 
Print["Tensor dimension ",type[tens[t]][[3]], " is not the manifold dimension", di," !"]; Return[]];
If[Not[MemberQ[Tuples[Range[di],pt],ii]],Print["Invalid argument ",ii, " !"];Return[]];
If[Not[MemberQ[Tuples[Range[di],qt],ji]],Print["Invalid argument ",ji, " !"];Return[]];
v = D[cu[t],t];xt =cu[t]/.List->Sequence;
wv[i_,j_]:=Sum[\[CapitalGamma][j,s,i][xt]*v[[s]],{s,di}];
(* The components *)
dt = D[tens[t][ii,ji],t];
dti =Sum[Sum[tens[t][Join[ii[[1;;l-1]],{s},ii[[l+1;;]]],ji]*wv[ii[[l]],s],{s,di}],{l,pt}];
dtj = Sum[Sum[tens[t][ii,Join[ji[[1;;l-1]],{s},ji[[l+1;;]]]]*wv[s,ji[[l]]],{s,di}],{l,qt}]; 
dt-dti+dtj];

tensorD1[\[CapitalGamma]_,tens_][cu_,t_]:= Module[{di,pt,qt},
(* Syntax control *)
If[Head[cu[t]]===List, di = Length[cu[t]],Print[cu[t]," is not a curve!"];Return[]];
If[Head[tens[t]]===tensor, pt = type[tens[t]][[2,1]];qt= type[tens[t]][[2,2]],
Print[tens[t]," is not a tensor!"];Return[]];
tensor[absD1[\[CapitalGamma],tens][cu,t],pt,qt,di]];


(* ::Subsubsection:: *)
(*The torsion tensor*)


tors[\[CapitalGamma]_][x__] := Module[{di = Length[{x}],to,i,j,k},
to[{i_,j_},{k_}]:=\[CapitalGamma][k,i,j][x]-\[CapitalGamma][k,j,i][x];
tensor[to,2,1,di]];


(* ::Subsubsection:: *)
(*The curvature tensor*)


RC[\[CapitalGamma]_][x__][k_,i_,h_,j_]:=D[\[CapitalGamma][k,j,i][x],{x}[[h]]]- D[\[CapitalGamma][k,h,i][x],{x}[[j]]]+
Sum[(\[CapitalGamma][k,h,l][x]\[CapitalGamma][l,j,i][x]-\[CapitalGamma][k,j,l][x]\[CapitalGamma][l,h,i][x]),{l,Length[{x}]}];

RCtensor[\[CapitalGamma]_][x__]:=Module[{arr,di=Length[{x}],k,h,i,j},
arr[{h_,i_,j_},{k_}]:=RC[\[CapitalGamma]][x][k,h,i,j];tensor[arr,3,1,di]];

Riccitensor[\[CapitalGamma]_][x__] := contraction[RCtensor[\[CapitalGamma]][x],2,1];

rct[\[CapitalGamma]_][x__][i_,j_]:= Sum[RC[\[CapitalGamma]][x][h,i,h,j],{h,Length[{x}]}];


(* ::Subsection:: *)
(*Riemannian spaces*)


(* ::Subsubsection:: *)
(*A general symmetric, twofold covariant tensor field*)


RG[gg_:g][x__]:=Normal[SymmetrizedArray[{i_,j_}:>gg[x][i,j],{Length[{x}],Length[{x}]},
{Symmetric[{1,2}]}]];

rg[gg_:g][x__][i_,j_] :=Evaluate[RG[gg][x]][[i,j]];

(* genRG[x__]:=RG[{x}];

specRG[h_][x__]:=RG[{x},h]; *)

RGtensor[gg_:g][x__]:=ArrayTotensor[RG[gg][x],2];


(* ::Subsubsection::Closed:: *)
(*Inverse metric tensor*)


INV[xx_][u__]:= Module[{m= Length[{u}]},
If[MatrixQ[xx[u]], Return[Inverse[xx[u]]]];
If[MatrixQ[xx[{u}]], Return[Inverse[xx[{u}]]]];
 Print["Wrong argument!"];Return[]];

inv[xx_][u__][i_,j_]:= INV[xx][u][[i,j]];


(* ::Subsubsection:: *)
(*The Pseudo-Euclidean spaces*)


eucl[u__][i_,j_] := KroneckerDelta[i,j];

EUCL[x__]:= RG[eucl][x];

lor[i_,j_,k_:indd]:=Module[{},If[Not[IntegerQ[i]]||Not[IntegerQ[j]]||(i<=0)||(j<=0),
Print["Invalid arguments!"];Return[]];
If[(i==j)&&(i<=k),-1,If[(i==j)&&(i>k),1,If[(i!=j),0]]]];

lorg[x__][i_,j_,k_:indd]:= lor[i,j,k];

 RGposk[r_][s__]:= Module[{n=Length[{s}],i,j,f,t,diag},
f[i_][t__] :=r^2* Product[(Cos[{t}[[j]]])^2,{j,i+1,n}];
diag [t__]:= Table[f[i][t],{i,1,n}]; 
DiagonalMatrix[diag[s]]];

rgposk[r_][s__][i_,j_]:= RGposk[r][s][[i,j]];



(* ::Subsubsection:: *)
(*The inner metric of an immersion*)


rgt[f_][x__][i_,j_] := tv[f][i][x].tv[f][j][x];

RGT[f_][x__]:= Table[rgt[f][x][i,j],{i,1,Length[{x}]},{j,1,Length[{x}]}];

dV[ff_]:= Module[{m,i,j},m = Length[{##}];Sqrt[Det[RGT[ff][##]]]]&;

dVol[\[Gamma]\[Gamma]_]:= Module[{m,i,j},m = Length[{##}];Sqrt[Abs[Det[Table[\[Gamma]\[Gamma][##][i,j],{i,m},{j,m}]]]]]&;


(* ::Subsubsection:: *)
(*The Christoffel symbols*)


chr[j_,k_,i_,g_:rg[]][x__]:=
(D[g[x][i,j],{x}[[k]]] +D[g[x][j,k],{x}[[i]]] -D[g[x][k,i],{x}[[j]]])/2;

\[CapitalGamma]g[g_:rg[]][k_,i_,j_][x__] := Module[{invg,gg,dd = Length[{x}],h,l,u},
If[Not[MemberQ[Tuples[Range[Length[{x}]],3],{k,i,j}]],
Print["Invalid indices ", k,",",i,",",j]; Return[]];
gg:=Table[g[x][h,l],{h,dd},{l,dd}];
invg = Inverse[gg];
Sum[invg[[k,h]]*chr[h,i,j,g][x], {h,dd}]];

\[CapitalGamma]f[ff_][h_,i_,j_][u__]:= \[CapitalGamma]g[rgt[ff]][h,i,j][u];

\[CapitalGamma]0[i_,j_,k_][x__] := 0;


(* ::Subsubsection:: *)
(*The covariant curvature tensor*)


covRCtensor[gg_:rg[]][x__]:= contraction[RGtensor[gg][x]\[CircleTimes]RCtensor[\[CapitalGamma]g[gg]][x],2,1];

covrc[gg_:rg[]][k_,h_,i_,j_][x__]:=covRCtensor[gg][x][{k,h,i,j},{}];

sectionalcurvature[gg_:rg[]][x__][X_, Y_]:=Module[{di=Length[{x}],tX,tY,gh,crt,nn,zz,i,j,k,l},
If[(Head[X]=== List)&&(Length[X]==di), tX=vecTotensor[X],
Print["Invalid argument X = ",X];Return[]];
If[(Head[Y]=== List)&&(Length[Y]==di), tY=vecTotensor[Y],
Print["Invalid argument Y = ",Y]; Return[]];
gh[i_,j_]:=gg[x][i,j];nn=tensorscpr[gh][tX\[Wedge]tY,tX\[Wedge]tY];
crt =covRCtensor[gg][x];
zz=Sum[crt[{i,j,k,l},{}]*X[[i]]*Y[[j]]*X[[k]]*Y[[l]],{i,di},{j,di},{k,di},{l,di}];
2*zz/nn];


(* ::Subsubsection::Closed:: *)
(*Tensor Metric*)


tensorscpr[g_][ta_,tb_]:= Module[{pp,qq,nn,h,gg,hh,mg,i,j,s,t,tt,tta,ttb,tfa,tfb},
If[Not[(Head[ta]===tensor)&&(Head[tb] ===tensor)], 
Print["Arguments must be tensors!"];Return[]];
If[(type[ta][[2]]===type[tb][[2]])&&(type[ta][[3]]===type[tb][[3]]),
pp=type[ta][[2,1]];qq=type[ta][[2,2]];nn=type[ta][[3]],
Print["tensors belong to different spaces!"];Return[]];
If[pp==0,tta = ta;ttb=tb;Goto[SUM]];
mg=Table[g[i,j],{i,nn},{j,nn}];
h[i_,j_] :=Inverse[ mg][[i,j]];
hh[i_List,j_List]:=Product[h[i[[s]],j[[s]]],{s,pp}];
tfa[{},i_List] :=Sum[hh[i[[1;;pp]],tpl[nn,pp][[s]]]*ta[tpl[nn,pp][[s]],i[[pp+1;;pp+qq]]],
{s,nn^pp}];
tfb[{},i_List] :=Sum[hh[i[[1;;pp]],tpl[nn,pp][[s]]]*tb[tpl[nn,pp][[s]],i[[pp+1;;pp+qq]]],
{s,nn^pp}];
tta =tensor[tfa,0,pp+qq,nn];ttb = tensor[tfb,0,pp+qq,nn];
Label[SUM];
gg[i_List,j_List]:=Product[g[i[[s]],j[[s]]],{s,pp+qq}];
Sum[gg[tpl[nn,pp+qq][[t]],tpl[nn,pp+qq][[tt]]]*
tta[{},tpl[nn,pp+qq][[t]]]*ttb[{},tpl[nn,pp+qq][[tt]]],{t,nn^(pp+qq)},{tt,nn^(pp+qq)}]];

tensorfieldscpr[gg_:rg[]][ta_,tb_][x__]:=Module[{gh,i,j},
gh[i_,j_]:=gg[x][i,j];tensorscpr[gh][ta,tb]];


(* ::Subsubsection:: *)
(*The n - sphere. Elliptic Riemannian geometry*)


spher[r_][s__]:=r*unit[s];

sph[r_][x__] := r*invsterproj[{x}];

RGpos[r_][s__] := FullSimplify[RGT[spher[r]][s], Element[{s},Reals]];

RGcnfpos[r_][x__]:= (((2*r)/(1+ {x}.{x}))^2)*EUCL[x];

rgcnfpos[r_][i_,j_][x__]:= RGcnfpos[r][x][[i,j]]; 


(* ::Subsubsection:: *)
(*Hyperbolic Riemannian geometry*)


hypsp[r_][s__]:= Module[{n=Length[{s}],t},
If[n==1, Return[{r Cosh[s],r Sinh[s]}]];
t=Table[{s}[[i]],{i,2,n}];
r*Join[{Cosh[{s}[[1]]]},Evaluate[Sinh[{s}[[1]]]*unit[t]]]];

RGhypsp[r_][s__] := Module[{n=Length[{s}],i,j,rm},
rm:=FullSimplify[Table[pssp[tv[hypsp[r]][i][s],tv[hypsp[r]][j][s]],{i,n},{j,n}],
Element[{s},Reals]];
rm];

rghypsp[r_][s__][i_,j_]:= RGhypsp[r][s][[i,j]];

hypcnf[r_][s__]:= r*Join[{(1+{s}.{s})/(1-{s}.{s})},2*{s}/(1-{s}.{s})];

RGhypcnf[r_][s__] := Module[{n=Length[{s}],i,j,rm},
rm:=FullSimplify[Table[pssp[tv[hypcnf[r]][i][s],tv[hypcnf[r]][j][s]],{i,n},{j,n}],
Element[{s},Reals]];rm];

rghypcnf[r_][s__][i_,j_]:= RGhypcnf[r][s][[i,j]];

RGneg[r_][s__] := (2*r/(-1+{s}.{s}))^2*IdentityMatrix[Length[{s}]];

rgneg[r_][s__][i_,j_]:= RGneg[r][s][[i,j]];


(* ::Subsection:: *)
(*The End*)


End[]


Protect["`*"];

EndPackage[]




